'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _constants = require('../constants');

var MediaSharesUtils = {};

/**
 * parses the relevant values for mediaShares: contentId, disposition
 * @param {Object} mediaShares
 * @returns {undefined}
 */
MediaSharesUtils.parse = function (mediaShares) {
  if (mediaShares) {
    return {
      contentId: MediaSharesUtils.getContentId(mediaShares),
      disposition: MediaSharesUtils.getDisposition(mediaShares)
    };
  }

  return null;
};

/**
 * get the previous and current mediaShares values parsed, as well as the boolean updates
 * @param {Object} oldShare
 * @param {Object} newShare
 * @returns {Object}
 * previous: {Object} old share, current: {Object} new share,
 */
MediaSharesUtils.getMediaShares = function (oldShare, newShare) {
  var previous = oldShare && MediaSharesUtils.parse(oldShare);
  var current = newShare && MediaSharesUtils.parse(newShare);

  return {
    previous: previous,
    current: current
  };
};

/**
 * get the floor disposition (released, granted)
 * @param {Object} mediaShares
 * @returns {Boolean} disposition
 */
MediaSharesUtils.getDisposition = function (mediaShares) {
  var contentFloor = MediaSharesUtils.extractContentFloor(mediaShares);

  return contentFloor ? contentFloor.disposition : null;
};

/**
 * extract the content property from media shares
 * @param {Object} mediaShares
 * @returns {Object}
 */
MediaSharesUtils.extractContent = function (mediaShares) {
  if (!mediaShares || !mediaShares.length) {
    return null;
  }

  return mediaShares.find(function (share) {
    return share.name === _constants.CONTENT;
  }) || null;
};

/**
 * extract the floor property from content object
 * @param {Object} content
 * @returns {Object}
 */
MediaSharesUtils.extractFloor = function (content) {
  if (!content) {
    return null;
  }

  return content.floor;
};

/**
 * extract the content's floor from media shares
 * @param {Object} mediaShares
 * @returns {Object}
 */
MediaSharesUtils.extractContentFloor = function (mediaShares) {
  var content = MediaSharesUtils.extractContent(mediaShares);

  return MediaSharesUtils.extractFloor(content);
};

/**
 * get who is sharing from media shares
 * @param {Object} mediaShares
 * @returns {Object}
 */
MediaSharesUtils.getContentId = function (mediaShares) {
  var contentFloor = MediaSharesUtils.extractContentFloor(mediaShares);

  if (!contentFloor || !contentFloor.beneficiary) {
    return null;
  }

  return contentFloor.beneficiary.id;
};

exports.default = MediaSharesUtils;
//# sourceMappingURL=mediaSharesUtils.js.map
